# -*- coding: utf-8 -*-
# Copyright (c) Polyconseil SAS. All rights reserved.

from __future__ import absolute_import
from __future__ import unicode_literals

import contextlib
import functools
import io
import json
import os.path
import shutil
import subprocess  # noqa: S404
import tempfile
import time

import jinja2
import pkg_resources
import yaml

from .six import configparser


def copy_resource(resource, destination, package='grocker'):
    resource_path = pkg_resources.resource_filename(package, resource)
    if pkg_resources.isdir(resource_path):
        shutil.copytree(resource_path, destination)
    else:
        shutil.copy2(resource_path, destination)


def load_yaml(file_path):
    with io.open(file_path, encoding='utf-8') as fp:
        return yaml.safe_load(fp.read())


def dump_yaml(file_path, data):
    directory = os.path.dirname(file_path)
    if not os.path.exists(directory):
        os.makedirs(directory)

    with io.open(file_path, 'w') as fp:
        return yaml.safe_dump(data, stream=fp, indent=True)


def load_yaml_resource(resource, package='grocker'):
    resource_path = pkg_resources.resource_filename(package, resource)
    return load_yaml(resource_path)


def render_template(template_path, output_path, context):
    env = jinja2.Environment()  # noqa: S701
    env.filters['jsonify'] = json.dumps

    with io.open(template_path, encoding='utf-8') as template_file:
        template = env.from_string(template_file.read())

    output = template.render(**context)

    with io.open(output_path, mode='w', encoding='utf-8') as output_file:
        output_file.write(output)


@contextlib.contextmanager
def pip_conf(pip_conf_path=None):
    """
    Use or fake (when it does not exist) a pip config file.

    Args:
        pip_conf_path: str, pip configuration file to use if it exists

    Yield:
        str, the path to the pip configuration file (or the faked one)
    """
    if pip_conf_path is None or not os.path.exists(pip_conf_path):
        with tempfile.NamedTemporaryFile('w', dir=os.path.expanduser('~/.cache')) as f:
            config = configparser.RawConfigParser()
            config.add_section('global')
            for key in ['index-url', 'timeout', 'extra-index']:
                try:
                    output = subprocess.check_output(  # noqa: S603,S607
                        ['pip', 'config', 'get', 'global.{}'.format(key)],
                    )
                    value = output.decode().strip()
                    config.set('global', key, value)
                except subprocess.CalledProcessError:
                    pass
            config.write(f)
            f.flush()
            yield f.name
    else:
        yield pip_conf_path


def retry(exception, tries=3, delay=1):
    def decorator(function):
        @functools.wraps(function)
        def inner(*args, **kwargs):
            for i in range(tries):
                try:
                    return function(*args, **kwargs)
                except exception:
                    if i + 1 == tries:
                        raise
                    time.sleep(delay)
        return inner
    return decorator


def get_version_from_requirement(requirement):
    if len(requirement.specifier) != 1:
        raise ValueError("Only exact specifier are accepted: %s" % requirement)
    spec = list(requirement.specifier)[0]
    if spec.operator not in ('==', '==='):
        raise ValueError("Only exact specifier are accepted: %s" % requirement)
    return spec.version



# -*- coding: utf-8 -*-
# Copyright (c) Polyconseil SAS. All rights reserved.

from __future__ import absolute_import
from __future__ import unicode_literals

import pkg_resources

__version__ = pkg_resources.get_distribution('grocker').version
__copyright__ = '2015, Polyconseil'



# -*- coding: utf-8 -*-
# Copyright (c) Polyconseil SAS. All rights reserved.

from __future__ import absolute_import
from __future__ import unicode_literals

from .. import __version__
from .. import utils


def image_name(config, role):
    image_name_template = 'grocker-{runtime}-{role}:{version}-{hash}'
    if role == 'wheel-server':
        image_name_template = 'grocker-{role}:{version}'

    if config['docker_image_prefix']:
        image_name_template = '{prefix}/' + image_name_template

    return image_name_template.format(
        prefix=config['docker_image_prefix'],
        runtime=config['runtime'].replace('/', '-'),
        role=role,
        version=__version__,
        hash=utils.config_identifier(config),
    )


def wheel_volume_name(config):
    return 'grocker-wheel-cache-{version}-{runtime}-{hash}'.format(
        version=__version__,
        runtime=config['runtime'].replace('/', '-'),
        hash=utils.config_identifier(config),
    )



# -*- coding: utf-8 -*-
# Copyright (c) Polyconseil SAS. All rights reserved.

from __future__ import absolute_import
from __future__ import unicode_literals

import hashlib
import os.path

import docker
import pkg_resources

from . import helpers

GROUP_SEPARATOR = b'\x1D'
RECORD_SEPARATOR = b'\x1E'
UNIT_SEPARATOR = b'\x1F'


def config_identifier(config):
    """
    Hash config to get an unique identifier.

    Args:
        config (dict): Grocker config

    Returns:
        str: Config identifier (SHA 256)

    """
    def unit_list(l):
        return UNIT_SEPARATOR.join(sorted(x.encode('utf-8') for x in l))

    dependencies = unit_list(get_dependencies(config, with_build_dependencies=True))
    repositories = RECORD_SEPARATOR.join(
        unit_list([name] + [cfg[x] for x in sorted(cfg)])
        for name, cfg in config['repositories'].items()
    )
    data = GROUP_SEPARATOR.join([
        dependencies,
        repositories,
    ])
    digest = hashlib.sha256(data)
    return digest.hexdigest()


def default_image_name(config, release):
    req = pkg_resources.Requirement.parse(release)
    if not str(req.specifier).startswith('=='):
        raise RuntimeError("Only fixed version can use default image name.")

    docker_image_prefix = config['docker_image_prefix']
    if config['image_base_name']:
        img_name = config['image_base_name']
    elif req.extras:
        img_name = "{project}-{extra_requirements}".format(
            project=req.project_name,
            extra_requirements='-'.join(req.extras),
        )
    else:
        img_name = req.project_name
    img_name += ":{project_version}".format(
        project_version=str(req.specifier)[2:],
    )
    return '/'.join((docker_image_prefix, img_name)) if docker_image_prefix else img_name


def docker_get_client(min_version=None):
    client = docker.from_env()
    if min_version and client.version()['ApiVersion'].split('.') <= min_version.split('.'):
        raise RuntimeError(
            'Docker API version should be at least {expected} ({current})'.format(
                current=client.version()['ApiVersion'],
                expected=min_version,
            ),
        )
    return client


def get_dependencies(config, with_build_dependencies=False):
    runtime = config['runtime']
    runtime_dependencies = config['runtimes'][runtime]['dependencies']

    dependencies = (
        runtime_dependencies.get('run', [])
        + config['dependencies'].get('run', [])
    )

    if with_build_dependencies:
        dependencies += (
            runtime_dependencies.get('build', [])
            + config['dependencies'].get('build', [])
        )

    return dependencies


def parse_config(config_paths, **kwargs):
    """
    Generate config regarding precedence order.

    Precedence order is defined as :

    1. Command line arguments
    2. project ``.grocker.yml`` file (or the one specified on the command line)
    3. the grocker ``resources/grocker.yaml`` file
    """
    config = helpers.load_yaml_resource('resources/grocker.yaml')

    if not config_paths and os.path.exists('.grocker.yml'):
        config_paths = ['.grocker.yml']

    for config_path in config_paths:
        project_config = helpers.load_yaml(config_path)
        config.update(project_config or {})

    config.update({k: v for k, v in kwargs.items() if v})

    return config



# -*- coding: utf-8 -*-
# Copyright (c) Polyconseil SAS. All rights reserved.

from __future__ import absolute_import
from __future__ import unicode_literals

import base64
import logging
import zlib

from . import naming
from . import op
from .. import six
from .. import utils

logger = logging.getLogger(__name__)


def get_pip_env(pip_conf):
    def get(cfg, section, option, default=None):
        try:
            return cfg.get(section, option)
        except (six.configparser.NoSectionError, six.configparser.NoOptionError):
            return default

    if not pip_conf:
        return {}

    logger.info('-> Pip use configuration from %s.', pip_conf)
    config = six.configparser.ConfigParser()
    config.read(pip_conf)

    env = {
        'PIP_INDEX_URL': get(config, 'global', 'index-url'),
        'PIP_EXTRA_INDEX_URL': get(config, 'global', 'extra-index-url'),
        'PIP_TIMEOUT': get(config, 'global', 'timeout'),
        'PIP_TRUSTED_HOST': get(config, 'global', 'trusted-host'),
    }
    env = {k: v for k, v in env.items() if v}
    logger.debug('pip using env: %s', env)
    return env


def compile_wheels(docker_client, config, release, pip_conf):
    wheels_destination_volume = op.get_or_create_data_volume(
        docker_client,
        naming.wheel_volume_name(config),
        role='wheel',
        labels={
            'grocker.runtime': config['runtimes'][config['runtime']]['runtime'],
            'grocker.config.hash': utils.config_identifier(config),
        },
    )
    volumes = {
        wheels_destination_volume.name: {
            'bind': '/home/grocker/packages',
            'mode': 'rw',
        },
    }
    command = ['--python', config['runtimes'][config['runtime']]['runtime'], release]
    environment = get_pip_env(pip_conf)

    if config['pip_constraint']:
        with open(config['pip_constraint'], 'rb') as fp:
            constraints = fp.read()
        environment['PIP_CONSTRAINT_CONTENT'] = base64.b64encode(zlib.compress(constraints)).decode()

    return op.docker_run_container(
        docker_client,
        naming.image_name(config, 'compiler'),
        command,
        volumes=volumes,
        environment=environment,
    )



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) Polyconseil SAS. All rights reserved.

from __future__ import absolute_import
from __future__ import unicode_literals

import logging

import click

from . import __version__
from . import builders
from . import cleanners
from . import helpers
from . import loggers
from . import utils

logger = logging.getLogger('grocker')


@click.group()
@click.version_option(__version__)
@click.option('-v', '--verbose', count=True)
def main(verbose):
    loggers.setup(verbose > 0)


@main.command()
@click.option('-a', '--all-versions/--only-old-versions', default=False)
@click.option('-f', '--including-final-images/--excluding-final-images', default=False)
def purge(all_versions, including_final_images):
    """Purge Grocker created Docker stuff."""
    docker_client = utils.docker_get_client()
    cleanners.docker_purge_container(docker_client, current_version=all_versions)
    cleanners.docker_purge_volumes(docker_client, current_version=all_versions)
    cleanners.docker_purge_images(docker_client, current_version=all_versions, runner=including_final_images)


@main.command()
@click.option(
    '-c', '--config', multiple=True, type=click.Path(exists=True), metavar='<filename>',
    help='Grocker config file',
)
@click.option('-r', '--runtime', metavar='<runtime>', help="runtime used to build and run this image")
@click.option(
    '--pip-conf', type=click.Path(exists=True), metavar='<filename>',
    help="pip configuration file used to download dependencies (by default use pip config getter)",
)
@click.option(
    '--pip-constraint', type=click.Path(exists=True), metavar='<filename>',
    help="pip constraint file used to download dependencies",
)
@click.option('-e', '--entrypoint', metavar='<entrypoint>', help="Docker entrypoint to use to run this image")
@click.option('--volume', multiple=True, metavar='<volume>', help="Container storage and configuration area")
@click.option('--port', multiple=True, metavar='<port>', help="Port on which a container will listen for connections")
@click.option('--env', multiple=True, metavar='<env=value>', help="Additional environment variable for final image")
@click.option(
    '--image-prefix', metavar='<uri>',
    help='docker registry or account on Docker official registry to use',
)
@click.option(
    '--image-base-name', metavar='<name>',
    help="base name for the image (eg '<image-prefix>/<image-base-name>:<image-version>')",
)
@click.option('-n', '--image-name', metavar='<name>', help="name used to tag the build image")
@click.option(
    '--result-file', type=click.Path(exists=False), metavar='<filename>',
    help="yaml file where results (image name, ...) are written",
)
@click.option(
    '--build-dependencies/--no-build-dependencies', default=True,
    help='build the dependencies',
)
@click.option(
    '--build-image/--no-build-image', default=True,
    help='build the docker image',
)
@click.option(
    '--push/--no-push', default=True,
    help='push the image',
)
@click.argument('release')
def build(release, build_dependencies, build_image, push, **kwargs):
    """Build docker image for <release> (version specifiers can be used)."""
    collect = {}  # will contain all collected information
    docker_client = utils.docker_get_client()
    collect['release'] = release

    config = utils.parse_config(
        kwargs['config'],
        runtime=kwargs['runtime'],
        entrypoint_name=kwargs['entrypoint'],
        pip_constraint=kwargs['pip_constraint'],
        docker_image_prefix=kwargs['image_prefix'],
        image_base_name=kwargs['image_base_name'],
        volumes=kwargs['volume'],
        ports=kwargs['port'],
        envs=dict(item.split('=', 1) for item in kwargs['env']),
    )
    image_name = kwargs['image_name'] or utils.default_image_name(config, release)
    collect['image'] = image_name

    # Raise if grocker do not known the runtime
    if config['runtime'] not in config['runtimes']:
        raise RuntimeError('Unknown runtime: %s', config['runtime'])

    if config['runtimes'][config['runtime']].get('deprecated'):
        logging.warning(
            "Runtime %s is deprecated, please update to more recent runtime",
            config['runtime'],
        )

    if build_dependencies:
        logger.info('Compiling dependencies...')
        builders.get_or_build_root_image(docker_client, config)
        compiler = builders.get_or_build_compiler_image(docker_client, config)
        collect['compiler_image'] = compiler.tags[0]

        with helpers.pip_conf(pip_conf_path=kwargs['pip_conf']) as pip_conf:
            builders.compile_wheels(
                docker_client=docker_client,
                config=config,
                release=release,
                pip_conf=pip_conf,
            )

    if build_image:
        logger.info('Building image...')
        root_image = builders.get_or_build_root_image(docker_client, config)
        collect['root_image'] = root_image.tags[0]
        builders.build_runner_image(
            docker_client=docker_client,
            config=config,
            name=image_name,
            release=release,
        )

    if push:
        if not builders.is_prefixed_image(image_name):
            logger.warning('Not pushing any image since the registry is unclear in %s', image_name)
        else:
            logger.info('Pushing image...')
            image = builders.docker_push_image(docker_client, image_name)
            collect['hash'] = (
                builders.get_manifest_digest(image_name)
                or [x.split('@')[1] for x in image.attrs['RepoDigests']][0]
            )

    if kwargs['result_file']:
        helpers.dump_yaml(kwargs['result_file'], collect)


if __name__ == '__main__':
    main()



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) Polyconseil SAS. All rights reserved.

from __future__ import absolute_import
from __future__ import unicode_literals

import shutil
import tempfile

# pylint: disable=unused-import,import-error
try:
    import configparser
except ImportError:
    import ConfigParser as configparser  # noqa
# pylint: enable=unused-import,import-error


class TemporaryDirectory(object):
    """Create and return a temporary directory.

    This has the same behavior as mkdtemp but can be used as a context manager.
    For example:

        with TemporaryDirectory() as tmpdir:
            ...

    Upon exiting the context, the directory and everything contained
    in it are removed.
    """

    def __init__(self, suffix="", prefix="tmp", directory=None):
        self.name = tempfile.mkdtemp(suffix, prefix, directory)

    def __repr__(self):
        return "<{} {!r}>".format(self.__class__.__name__, self.name)

    def __enter__(self):
        return self.name

    def __exit__(self, exc, _value, _tb):
        shutil.rmtree(self.name)



# -*- coding: utf-8 -*-
# Copyright (c) Polyconseil SAS. All rights reserved.

from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import contextlib
import logging
import os.path

import docker
import docker.errors
import docker.utils.json_stream
import requests
import requests.exceptions

from .. import __version__
from .. import helpers
from .. import six

logger = logging.getLogger(__name__)


def is_prefixed_image(name):
    return '/' in name


@contextlib.contextmanager
def docker_build_context(context_path):
    with six.TemporaryDirectory() as tmp_dir:
        build_dir = os.path.join(tmp_dir, 'build')
        helpers.copy_resource(context_path, build_dir)
        yield build_dir


def docker_build_image(docker_client, path, name, role=None, labels=None, **kwargs):
    computed_labels = {
        'grocker.version': __version__,
        'grocker.image.role': role,
    }
    computed_labels.update(labels or {})
    stream = docker_client.api.build(
        path=path,
        tag=name,
        rm=True,
        forcerm=True,
        labels=computed_labels,
        **kwargs
    )
    _inspect_stream(stream)
    try:
        return docker_client.images.get(name)
    except docker.errors.ImageNotFound:
        raise RuntimeError('Image build failed')


def docker_get_or_build_image(docker_client, name, builder):
    try:
        return docker_client.images.get(name)
    except (requests.exceptions.HTTPError, docker.errors.ImageNotFound):
        try:
            return docker_pull_image(docker_client, name)
        except (requests.exceptions.HTTPError, docker.errors.NotFound):
            image = builder(docker_client)
            if is_prefixed_image(name):
                image = docker_push_image(docker_client, name)
            return image


def get_or_create_data_volume(docker_client, name, role, labels=None):
    logger.info('Creating volume %s...', name)
    computed_labels = {
        'grocker.version': __version__,
        'grocker.image.role': role,
    }
    computed_labels.update(labels or {})
    return docker_client.volumes.create(
        name=name,
        labels=computed_labels,
    )


def docker_pull_image(docker_client, name):
    logger.info('Pulling image %s...', name)
    return docker_client.images.pull(name)


def docker_push_image(docker_client, name):
    logger.info('Pushing image %s...', name)
    docker_client.images.push(name)
    return docker_client.images.get(name)


def get_manifest_digest(name):
    registry_repository, tag = name.split(':', 1)
    if '.' not in registry_repository:  # Docker Hub
        return None  # Docker HUB API is not documented

    registry, repository = registry_repository.split('/', 1)
    response = requests.head('https://{}/v2/{}/manifests/{}'.format(registry, repository, tag))
    return response.headers['Docker-Content-Digest']


def docker_run_container(docker_client, name, command, volumes=None, environment=None):
    logger.info(
        'Running %s on image %s (volumes:%s, environment:%s)',
        command, name, volumes, environment,
    )
    container = docker_client.containers.run(
        image=name,
        command=command,
        environment=environment,
        volumes=volumes,
        detach=True,
    )

    stream = container.attach(stream=True, logs=True)
    for line in stream:
        print(line.decode('utf-8'), end='')
    result = container.wait()
    container.remove()
    if result['StatusCode'] != 0:
        raise RuntimeError('Container exit with a non-zero return code (%s).' % result)


def _inspect_stream(stream):
    """Return some data about the stream."""
    error = False
    for line in docker.utils.json_stream.json_stream(stream):
        if 'stream' in line:
            print(line['stream'], end='')
        elif 'error' in line:
            print(line['error'])
            error = True
        else:
            print(line)
    print()
    return not error



# -*- coding: utf-8 -*-
# Copyright (c) Polyconseil SAS. All rights reserved.

from __future__ import absolute_import
from __future__ import unicode_literals

import contextlib
import logging
import os.path
import shutil

import docker.errors
from packaging import requirements

from . import naming
from . import op
from .. import __version__
from .. import helpers
from .. import utils

logger = logging.getLogger(__name__)


def build_root_image(docker_client, config):
    with op.docker_build_context('resources/docker/root-image') as build_dir:
        cfg = config['runtimes'][config['runtime']]
        context = {
            'base_image': cfg['image'],
            'repositories': config['repositories'],
            'runtime': cfg['runtime'],
            'grocker_version': __version__,
        }

        helpers.render_template(
            os.path.join(build_dir, 'Dockerfile.j2'),
            os.path.join(build_dir, 'Dockerfile'),
            context,
        )

        # FIXME(fbochu): Replace template by env vars
        helpers.render_template(
            os.path.join(build_dir, 'provision.sh.j2'),
            os.path.join(build_dir, 'provision.sh'),
            context,
        )

        dependencies = utils.get_dependencies(config)
        build_env = {
            'SYSTEM_DEPENDENCIES': ' '.join(dependencies),
        }
        return op.docker_build_image(
            docker_client,
            build_dir,
            naming.image_name(config, 'root'),
            buildargs=build_env,
            role='root',
        )


def build_compiler_image(docker_client, config):
    with op.docker_build_context('resources/docker/compiler-image') as build_dir:
        context = {
            'base_image': naming.image_name(config, 'root'),
            'runtime': config['runtimes'][config['runtime']]['runtime'],
        }

        helpers.render_template(
            os.path.join(build_dir, 'Dockerfile.j2'),
            os.path.join(build_dir, 'Dockerfile'),
            context,
        )

        dependencies = utils.get_dependencies(config, with_build_dependencies=True)
        build_env = {
            'SYSTEM_DEPENDENCIES': ' '.join(dependencies),
        }
        return op.docker_build_image(
            docker_client,
            build_dir,
            naming.image_name(config, 'compiler'),
            buildargs=build_env,
            role='compiler',
        )


def build_wheel_server_image(docker_client, config):
    with op.docker_build_context('resources/docker/wheel-server') as build_dir:
        return op.docker_build_image(
            docker_client,
            build_dir,
            naming.image_name(config, 'wheel-server'),
            role='wheel-server',
        )


def build_runner_image(docker_client, config, name, release):
    requirement = requirements.Requirement(release)

    # Markers would not make much sense here and url are unsupported.
    if requirement.marker or requirement.url:
        raise RuntimeError("Unsupported release specifier: %s" % release)

    with op.docker_build_context('resources/docker/runner-image') as build_dir:
        context = {
            'base_image': naming.image_name(config, 'root'),
            'entrypoint_name': config['entrypoint_name'],
            'app_name': requirement.name,
            'app_extras': ','.join(sorted(requirement.extras)),
            'app_version': helpers.get_version_from_requirement(requirement),
            'volumes': config['volumes'],
            'ports': config['ports'],
            'envs': config['envs'],
        }

        helpers.render_template(
            os.path.join(build_dir, 'Dockerfile.j2'),
            os.path.join(build_dir, 'Dockerfile'),
            context,
        )

        if config.get('pip_constraint'):
            shutil.copyfile(
                config['pip_constraint'],
                os.path.join(build_dir, 'constraints.txt'),
            )

        with wheel_server(docker_client, config) as wheel_server_ip:
            build_env = {
                'GROCKER_WHEEL_SERVER_IP': wheel_server_ip,
            }
            return op.docker_build_image(
                docker_client,
                build_dir,
                name,
                role='runner',
                buildargs=build_env,
                nocache=True,
            )


@contextlib.contextmanager
def wheel_server(docker_client, config):
    image = op.docker_get_or_build_image(
        docker_client,
        naming.image_name(config, 'wheel-server'),
        lambda client: build_wheel_server_image(client, config),
    )

    container = docker_client.containers.run(
        image=image.id,
        volumes={
            naming.wheel_volume_name(config): {
                'bind': '/wheels',
                'mode': 'ro',
            },
        },
        detach=True,
    )
    logger.info('Starting http server in container: %s', container.id)
    container.reload()
    server_ip = container.attrs['NetworkSettings']['IPAddress']
    logger.debug('http server running with ip: %s', server_ip)
    try:
        yield server_ip
    finally:
        helpers.retry(docker.errors.APIError)(container.remove)(force=True)
